# -*- coding: utf-8 -*-
"""
This module demonstrates the constrained optimization matching process utilizing a stripped down example. Nomenclature corresponds to the following paper for better orientation:
>>Constrained Optimization as Allocation Method in Local Flexibility Markets<< by Andreas Zeiselmair and Simon Köppl
Note: GLPK solver is required to run the script.

Created on Mon Jan 18 09:43:32 2021

@authors: JKowalczyk, VRegener, AZeiselmair
"""

import numpy as np
import pulp

# ---------------------------------------------
#   Function Description
# ---------------------------------------------


def eta(i, j, S_1, q):
    """Auxiliary function that indicates if the flex option features
    continuous power modulation (returns 0) or discrete power steps (returns 1)
    Args: Index Demand, Index Timestep, Set of percentages, Cardinality of S_1
    """
    S = list(range(0, q))
    c = sum([S_1[i][j][s] for s in S])
    if c < 0:
        return 0
    else:
        return 1


def delta(l, j, D):
    """Auxiliary function that indicates if the corresponding Demand
    is zero (returns 0) or not (returns 1)
    Args: Index Demand, Index Timestep, set of all Demands
    """
    if D[l][j] == 0:
        return 0
    else:
        return 1


def gamma(j, D, L):
    """Auxiliary function that indicates if the whole Demand of one timestep
    is zero (returns 0) or not (returns 1)
    Args: Index Timestep, Demand, Set of all Demands
    """
    if sum([delta(l, j, D) for l in L]) != 0:
        return 1
    else:
        return 0


# ---------------------------------------------
# Input Parameters
# ---------------------------------------------

# Number /Set of flex options
n = 3
I = range(n)
# Number /Set of time steps
m = 5
J = range(m)
J_2 = J[1:]
# Set of demands
L = range(1)
# Number / Set of discrete power steps (percentages)
q = 4
S_1 = {
    0: [[-1, -1, -1, -1] for j in J],
    1: [[-1, -1, -1, -1] for j in J],
    2: [[-1, -1, -1, -1] for j in J]
}
# Costs of flex option i at time period j
C_pos = {0: [10 for j in J], 1: [20 for j in J], 2: [5 for j in J]}
C_pos[2][3] = 100
C_neg = C_pos
# Flexibility demand at time period j
D = {0: [-3.16, -2.53, -1.59, -2.86, -2.74]}
# Penalty costs for non-fulfillment of demand
G = {0: [1000 for j in J]}
# Maximum number of calls of flex option per day
N = [96 for i in I]
# Total call duration per day
T_sum = [4, 96, 96]
# Maximum duration per call
T_max = [96, 2, 96]
# Minimum duration per call
T_min = [1, 1, 1]
# Minimum time between two calls
tau = [1, 1, 1]
# Remaining blocking time of previous day
tau_rem = [0, 0, 0]
# Maximum available power
P_max_pos = {0: [0 for j in J], 1: [0 for j in J], 2: [0 for j in J]}
P_max_neg = {0: [-40 for j in J], 1: [-30 for j in J], 2: [-50 for j in J]}
# Minimum available power
P_min_pos = {0: [0 for j in J], 1: [0 for j in J], 2: [0 for j in J]}
P_min_neg = {0: [0 for j in J], 1: [0 for j in J], 2: [0 for j in J]}
# Slope of the efficiency
e = {
    0: {
        0: [0.03, 0.03, 0.03],
        1: [0.03, 0.03, 0.03],
        2: [0.03, 0.03, 0.03],
        3: [0.03, 0.03, 0.03],
        4: [0.03, 0.03, 0.03],
        5: [0.03, 0.03, 0.03]
    }
}

# Auxiliary constant (large enough)
k = 100000
# Possible shares of power
S = range(4)

# ---------------------------------------------
# Transfer to pulp
# ---------------------------------------------
Matching = pulp.LpProblem("Matching", pulp.LpMinimize)

# Initialize Variables
P = pulp.LpVariable.dict('P', (I, J), cat='Continuous')
P_pos = pulp.LpVariable.dict('P_pos', (I, J), lowBound=0, cat='Continuous')
P_neg = pulp.LpVariable.dict('P_neg', (I, J), lowBound=0, cat='Continuous')

x = pulp.LpVariable.dict('x', (I, J), cat='Binary')
x_pos = pulp.LpVariable.dict('x_pos', (I, J), cat='Binary')
x_neg = pulp.LpVariable.dict('x_neg', (I, J), cat='Binary')

aux = pulp.LpVariable.dict('aux', (I, J_2), cat='Binary')

d_1 = pulp.LpVariable.dict('d_1', (J, L), lowBound=0, cat='Continuous')
d_2 = pulp.LpVariable.dict('d_2', (J, L), lowBound=0, cat='Continuous')

xi = pulp.LpVariable.dict('xi', (I, J, S), cat='Binary')
y = pulp.LpVariable.dict('y', (I, J, L), cat='Binary')
b = pulp.LpVariable.dict('b', (I, J, L), lowBound=0, cat='Continuous')

# Objective function
Matching += pulp.lpSum([
    C_pos[i][j] * P_pos[i, j] + C_neg[i][j] * P_neg[i, j] for i in I for j in J
]) + pulp.lpSum([(G[l][j] * (d_1[j, l] + d_2[j, l])) for j in J
                 for l in L]), "Z"

# Constraints
for l in L:
    for j in J:

        # Obtain Absolute Value of the missing Demand
        Matching += d_1[j, l] - d_2[j, l] <= delta(l, j, D) * (
            D[l][j] - sum([(e[l][j][i] * P[i, j])
                           for i in I]))
        Matching += delta(l, j, D) * (
            D[l][j] - sum([(e[l][j][i] * P[i, j])
                           for i in I])) <= d_1[j, l] - d_2[j, l]

for i in I:
    # If the activity begins at time step 0, the max activity time is T_max
    Matching += sum(
        [x[i, J[0] + k]
         for k in list(range(0, min(int(T_max[i]) + 1, m)))]) <= T_max[i]

    # If the activity begins at time step 0 then it has to be activated at least for T_min*4 TSs
    Matching += sum([
        x[i, J[0] + k] for k in list(range(0, min(int(T_min[i]), m)))
    ]) >= T_min[i] * x[i, J[0]]

    # Maximal Number of calls has to be respected
    Matching += (x[i, J[0]] + x[i, J[0] + m - 1] +
                 sum([aux[i, j] for j in J_2])) / 2 <= N[i]

    # Maximal call duration per day has to be respected
    Matching += sum([x[i, j] for j in J]) <= T_sum[i]

    for j in J:

        # Obtain P from P_pos and P_neg
        Matching += P[i, j] <= P_pos[i, j] - P_neg[i, j]
        Matching += -P[i, j] <= -(P_pos[i, j] - P_neg[i, j])

        # Obtain x-indicators
        # Obtain x_pos as indicator if P_pos >0 or not
        Matching += gamma(j, D, L) * x_pos[i, j] <= k * P_pos[i, j]
        Matching += gamma(j, D, L) * x_pos[i, j] * k >= P_pos[i, j]

        # Obtain x_neg as indicator if P_neg >0 or not
        Matching += gamma(j, D, L) * x_neg[i, j] <= k * P_neg[i, j]
        Matching += gamma(j, D, L) * x_neg[i, j] * k >= P_neg[i, j]

        # Restrict that either P_neg >0 or P_pos >0 or none holds
        Matching += x_neg[i, j] + x_pos[i, j] <= 1

        # Obtain x from x_pos and x_neg
        Matching += x_neg[i, j] + x_pos[i, j] <= x[i, j]
        Matching += x_neg[i, j] + x_pos[i, j] >= x[i, j]

        if j < tau_rem[i]:
            Matching += P[i, j] >= 0
            Matching += P[i, j] <= 0

        # If we have Percentages: we can choose only one
        Matching += eta(i, j, S_1,
                        q) <= eta(i, j, S_1, q) * sum([xi[i, j, s] for s in S])
        Matching += eta(i, j, S_1, q) * sum([xi[i, j, s]
                                             for s in S]) <= eta(i, j, S_1, q)

        Matching += eta(i, j, S_1, q) * P[i, j] <= eta(i, j, S_1, q) * (sum(
            [xi[i, j, s] * S_1[i][j][s]
             for s in S])) * (P_max_pos[i][j] + P_max_neg[i][j])

        Matching += eta(i, j, S_1, q) * (sum(
            [xi[i, j, s] * S_1[i][j][s]
             for s in S])) * (P_max_pos[i][j] + P_max_neg[i][j]) <= eta(
                 i, j, S_1, q) * P[i, j]

        #If P is continuous P_max has to be respected
        Matching += ((1 - eta(i, j, S_1, q)) * x_pos[i, j] * P_min_pos[i][j] <=
                     (1 - eta(i, j, S_1, q)) * P_pos[i, j])
        Matching += ((1 - eta(i, j, S_1, q)) * x_neg[i, j] * P_min_neg[i][j] >=
                     (1 - eta(i, j, S_1, q)) * (-P_neg[i, j]))
        Matching += ((1 - eta(i, j, S_1, q)) * P_pos[i, j] <=
                     (1 - eta(i, j, S_1, q)) * P_max_pos[i][j] * x_pos[i, j])
        Matching += ((1 - eta(i, j, S_1, q)) * (-P_neg[i, j]) >=
                     (1 - eta(i, j, S_1, q)) * P_max_neg[i][j] * x_neg[i, j])
        Matching += x_pos[i, j] * P_min_pos[i][j] <= P_pos[i, j]
        Matching += x_neg[i, j] * P_min_neg[i][j] >= (-P_neg[i, j])
        Matching += P_pos[i, j] <= P_max_pos[i][j] * x_pos[i, j]
        Matching += (-P_neg[i, j]) >= P_max_neg[i][j] * x_neg[i, j]

for i in I:
    for j in J_2:

        # Block Time
        for k in list(range(1, int(tau[i]) + 1)):
            if j - k >= J[0]:
                Matching += ((x[i, j] - x[i, j - 1]) + x[i, j - k]) <= 1

# Indicator if FO was activated/deactivated
        Matching += -aux[i, j] <= (x[i, j - 1] - x[i, j])
        Matching += (x[i, j - 1] - x[i, j]) <= aux[i, j]

        # If we activate a FO then it has to be deactivated at least after T_max TSs!
        Matching += (x[i, j] - x[i, j - 1]) + sum([
            x[i, j + k] for k in list(range(1, int(min(T_max[i] + 1, m - j))))
        ]) <= T_max[i]

        # If we activate a FO then it has to be activated at least for T_min*4 TSs
        for k in list(range(min(m - j, int(T_min[i])))):
            Matching += x[i, j + k] >= x[i, j] - x[i, j - 1]

# ---------------------------------------------
#   Matching
# ---------------------------------------------
print('Done preparing, start the matching...')

try:
    solver = pulp.GLPK_CMD()
    status = Matching.solve(solver)
    # ---------------------------------------------
    # Console Output
    # ---------------------------------------------
    headers = ["Time step", "Demand", "FO 1", "FO 2", "FO 3"]

    # Preallocate table
    data = np.zeros((m, len(headers)))
    for j in J:
        data[j, :] = [
            j, D[0][j], P[0, j].varValue, P[1, j].varValue, P[2, j].varValue
        ]

    # Format and print
    row_format = "{:>12}" * (len(headers))
    print(row_format.format(*headers))
    for header, row in zip(headers, data):
        print(row_format.format(*row))
except:
    print("Solver not Found! Please install GLPK or change Solver in line 260.")
